/*-----------------------------------------------------------------------------
  Project:  QRE
  File:     main.cpp

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Revision: 1.6 $
 *----------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/fs.h>
#include <nn/fnd.h>

#include <cstdio>
#include <cstring>

#include <mw/qre/qre.h>

#include "WBMHandler.h"
#include "GraphicsControl.h"

using namespace nn;
using fnd::ExpHeap;
using namespace mw::qre;

/* -------------------------------------------------------------------------- */

/** Reci* QRR[h̊R[hShift JISɌŒłBj */
const char* const QR_DATA  = "ЃACGXs[\nhttp://www.isp21.co.jp/";
/** ͉摜t@C */
const char *const QR_INPUT_FILENAME = "rom:/InputImage.bmp";
/** oQRR[hC[Wt@Cev[g */
const char* const QR_OUTPUT_FILENAME_TEMPLATE = "sdmc:/QR%02d.bmp";

/* -------------------------------------------------------------------------- */

/**
 * @brief RGB888摜̃oCgI[_[ϊϊB
 *
 * Windows BMP`RGB888摜̃oCgI[_[iBAGARjA
 * CTRvbgtH[RGB888̃oCgI[_[ɁiRAGABjϊ܂B
 *
 * @param[out] dst    CTR-RGB888摜i[obt@
 * @param[in]  src    WindowsBMP RGB888摜
 * @param[in]  dwidth 摜
 * @param[in]  height 摜̍
 * @param[in]  sig    XLC̃I[_[ (true: , false: )
 */
static void RGB888_WBM2CTR(u8 *dst, u8 *src, u16 dwidth, u16 height, bool sig)
{
    // MS Windows BMP1C̕4byteoE_
    u16 swidth = round4(dwidth * 3);

    int sy = 0;
    // MS Windows BMP̃XLC͉A
    // }CiX̏ꍇAォ牺ƂȂ
    if (sig)
        sy = height - 1;

    for (int dy = 0; dy < height; dy++) {
        int dty = dy * dwidth * 3;
        int sty = sy * swidth;
        for (int x = 0; x < dwidth; x++) {
            int tx = x * 3;
            dst[tx + dty] = src[tx + sty];
            dst[tx + dty + 1] = src[tx + sty + 1];
            dst[tx + dty + 2] = src[tx + sty + 2];
        }
        if (sig)
            sy--;
        else
            sy++;
    }
}

/**
 * @brief WindowsBMPǂݍށB
 *
 * w肳ꂽt@CBMP`̃t@Cǂݍ݁A
 * 摜\̂Ɋi[܂B
 *
 * @param[out] info    ǂݍ񂾉摜i[obt@
 * @param[in]  f_name  ǂݍރt@C
 * @param[in]  expHeap f̈mۂ邽߂̃q[vNX
 * @retval     true    I
 * @retval     false   s
 */
static bool ReadBMPFile(ImageInfo *info, const char *f_name, ExpHeap &expHeap)
{
    fs::FileInputStream fis;
    fis.Initialize(f_name);
    BitMapFileHeader bmfs(fis);
    if (bmfs.get_bits() != 24) {
        NN_LOG("Support only RGB24 bit image...\n");
        return false;
    }

    int header_size = bmfs.get_header_size();
    int data_size = bmfs.get_image_size();
    u8* ptr = reinterpret_cast<u8*>(expHeap.Allocate(data_size));
    if (!ptr) {
        NN_LOG("Memory allocation failed...\n");
        return false;
    }
    fis.Seek(header_size, fs::POSITION_BASE_BEGIN);
    fis.Read(ptr, data_size);
    fis.Finalize();

    s16 height = bmfs.get_height();
    bool sig = true;
    if (height < 0) {
        sig = false;
        height = -height;
    }

    info->width = bmfs.get_width();
    info->height = height;
    info->rgb888Data = reinterpret_cast<u8*>(expHeap.Allocate(info->width * info->height * 3));

    if (!info->rgb888Data) {
        NN_LOG("Memory allocation failed...\n");
        expHeap.Free(ptr);
        return false;
    }

    RGB888_WBM2CTR(info->rgb888Data, ptr, info->width, info->height, sig);
    expHeap.Free(ptr);
    return true;
}

/* -------------------------------------------------------------------------- */

/**
 * @brief QRGR[hCuTvvOC֐
 */
extern "C" void nnMain()
{
    /* vbgtH[̏ */
    // os::Initialize();
    fs::Initialize();
	nn::applet::CTR::Enable();

    /*
     * CuŎgp郁ݒ
     * ̎͂܂łAt@Ō̉łA
     * CuŎgp郁̊Ǘ@́Aǂ\
     */
    ExpHeap expHeap;
    expHeap.Initialize(os::GetDeviceMemoryAddress(),
                       os::GetDeviceMemorySize(),
                       os::ALLOCATE_OPTION_LINEAR);

    GraphicsControl graphics;
    graphics.InitializeGraphics(&expHeap);

    s32 fs_mem_size = fs::GetRomRequiredMemorySize(1, 1);
    void *fs_mem = expHeap.Allocate(fs_mem_size);
    Result fs_res = fs::MountRom(1, 1, fs_mem, fs_mem_size);
    if (fs_res.IsFailure()) {
        NN_LOG("ROM mount failed...\n");
        for (;;);
    }

    fs_res = fs::MountSdmc();
    if (fs_res.IsFailure()) {
        NN_LOG("SDMC mount failed...\n");
        for (;;);
    }

    ImageInfo img;
    if (!ReadBMPFile(&img, QR_INPUT_FILENAME, expHeap)) {
        NN_LOG("Image read failed...\n");
        for (;;);
    }

    QREncoder *encoder = new QREncoder;
    u32 data_size = std::strlen(QR_DATA);
    u32 memory_size = QREncoder::GetEncodeBufferSize(data_size, 3, 1);
    void *heap = expHeap.Allocate(memory_size);
    encoder->InitMemory(heap, memory_size);

    /* 쐬QRR[h̏ݒ */
    EncodeData info;
    /* o[W15Ƃ */
    info.version = 12;
    /* xH */
    info.ecc_level = ECC_LEVEL_H;
    /* Rec̏ */
    info.size = data_size;
    info.data = reinterpret_cast<u8*>(const_cast<char*>(QR_DATA));
    /*
     * RecTCYɍ킹ĕA
     * ̃Tvł͕قǑ傫ȃf[^͂邱Ƃ͑z肵ĂȂ
     */
    info.total = 0;
    /* Z̃TCYݒ */
    info.cell_size = 3;

    /* GR[hs */
    if (!encoder->EncodeWithImage(&info, &img)) {
        /* GR[hɎs */
        NN_LOG("QR encode failed...\n");
        expHeap.Free(img.rgb888Data);
        for (;;);
    }
    expHeap.Free(img.rgb888Data);

    int count = encoder->GetDivCount();
    for (int i = 0; i < count; i++) {

        int size = encoder->GetQRSize(i, true);
        if (!size) {
            /* f[^TCYꍇ͉Ȃ */
            NN_LOG("Data size illegal...\n");
            continue;
        }
        u8* bmp_data = reinterpret_cast<u8*>(expHeap.Allocate(size));
        if (!bmp_data) {
            NN_LOG("Memory allocation failed...\n");
            /* ̊mۂɎs */
            continue;
        }
        if (encoder->GetQRBMPData(bmp_data, size, i)) {

            char file_name[32];
            std::sprintf(file_name, QR_OUTPUT_FILENAME_TEMPLATE, i);
            fs::FileOutputStream fos;
            fos.Initialize(file_name, true);
            fos.SetSize(size);
            if (fos.Write(bmp_data, size, false) != size) {
                NN_LOG("File save failed...\n");
            }
            fos.Flush();
            fos.Finalize();
        }
        expHeap.Free(bmp_data);
    }

    int qr_size = encoder->GetQRSize(0, false);
    ImageInfo QRImage;
    QRImage.rgb888Data = reinterpret_cast<u8*>(expHeap.Allocate(qr_size));
    if (!QRImage.rgb888Data) {
        NN_LOG("Memory allocation failed...\n");
        for (;;);
    }
    encoder->GetQRData(&QRImage, 0);
    delete encoder;

    graphics.ClearDisplay(NN_GX_DISPLAY0);
    graphics.DrawImage(QRImage.rgb888Data, QRImage.width, QRImage.height);
    graphics.SwapBuffer();
    graphics.ClearDisplay(NN_GX_DISPLAY1);
    graphics.SwapBuffer();

    expHeap.Free(QRImage.rgb888Data);

    NN_LOG("Sample program finished.\n");

    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromDays(1));
}

/* -------------------------------------------------------------------------- */
